static char *rcs_id = "$Id: cisam.c,v 0.9 1994/10/31 07:10:44 mathias Exp $";
/*
 * cisam.c	- functions for cisamperl/rocisperl
 *
 * (C) 1993,1994 Mathias Koerber
 * please see accompanying file COPYRIGHT for details on
 * copyright and copying/(re-)distribution.
 *
 * This is an attempt at implementing a C-ISAM interface for perl.
 *
 * $Log: cisam.c,v $
 * Revision 0.9  1994/10/31  07:10:44  mathias
 * *** empty log message ***
 *
 */

#include "EXTERN.h"
#include "perl.h"
#include <isam.h>
#include <decimal.h>
#include <limits.h>

#ifdef READONLY
#define	ROMSG "This is read-only version of cisamperl"
#endif
#ifdef NODECIMAL
#define NODECMSG "Decimal type not implemented yet"
#endif
/*
 * reclen
 * for each file we keep the record length
 */

static int	isam_reclen[OPEN_MAX];

static enum uservars {
	UV_iserrno,
	UV_isrecnum,
	UV_isstat1,
	UV_isstat2
};

static enum usersubs {
/* c-isam file manipulation */
	US_isaddindex,
	US_isaudit,
	US_isbegin,
	US_isbuild,
	US_isclose,
	US_iscluster,
	US_iscommit,
	US_isdelcurr,
	US_isdelete,
	US_isdelindex,
	US_isdelrec,
	US_iserase,
	US_isflush,
	US_isindexinfo,
	US_islock,
	US_islogclose,
	US_islogopen,
	US_isopen,
	US_isread,
	US_isrecover,	
	US_isrelease,
	US_isrename,
	US_isrewcurr,
	US_isrewrec,
	US_isrewrite,
	US_isrollback,
	US_issetunique,
	US_isstart,
	US_isuniqueid,
	US_isunlock,
	US_iswrcurr,
	US_iswrite,
/* format conversion */
	US_ldchar,
	US_lddbl,
	US_lddblnull,
	US_lddecimal,
	US_ldfloat,
	US_ldfltnull,
	US_ldint,
	US_ldlong,
	US_stchar,
	US_stdbl,
	US_stdblnull,
	US_stdecimal,
	US_stfloat,
	US_stfltnull,
	US_stint,
	US_stlong,
/* decimal functions */
	US_deccvasc,
	US_dectoasc,
	US_deccvint,
	US_dectoint,
	US_deccvlong,
	US_dectolong,
	US_deccvflt,
	US_dectoflt,
	US_deccvdbl,
	US_dectodbl,
	US_decadd,
	US_decsub,
	US_decmul,
	US_decdiv,
	US_deccmp,
	US_deccopy,
	US_dececvt,
	US_decfcvt
};

static int usersub();
static int userset();
static int userval();

int
init_cisam()
{
    struct ufuncs uf;
    char *filename = "cisam.c";

    uf.uf_set = userset;
    uf.uf_val = userval;

#define MAGICVAR(name, ix) uf.uf_index = ix, magicname(name, &uf, sizeof uf)
    MAGICVAR("iserrno", UV_iserrno);
    MAGICVAR("isrecnum", UV_isrecnum);
    MAGICVAR("isstat1", UV_isstat1);
    MAGICVAR("isstat2", UV_isstat2);

/* c-isam file manipulation */
    make_usub("isaddindex",	US_isaddindex,	usersub, filename);
    make_usub("isaudit",	US_isaudit,	usersub, filename);
    make_usub("isaudit",	US_isaudit,	usersub, filename);
    make_usub("isbegin",	US_isbegin,	usersub, filename);
    make_usub("isbuild",	US_isbuild,	usersub, filename);
    make_usub("isclose",	US_isclose,	usersub, filename);
    make_usub("iscluster",	US_iscluster,	usersub, filename);
    make_usub("iscommit",	US_iscommit,	usersub, filename);
    make_usub("isdelcurr",	US_isdelcurr,	usersub, filename);
    make_usub("isdelete",	US_isdelete,	usersub, filename);
    make_usub("isdelindex",	US_isdelindex,	usersub, filename);
    make_usub("isdelrec",	US_isdelrec,	usersub, filename);
    make_usub("iserase",	US_iserase,	usersub, filename);
    make_usub("isflush",	US_isflush,	usersub, filename);
    make_usub("isindexinfo",	US_isindexinfo,	usersub, filename);
    make_usub("islock",		US_islock,	usersub, filename);
    make_usub("islogclose",	US_islogclose,	usersub, filename);
    make_usub("islogopen",	US_islogopen,	usersub, filename);
    make_usub("isopen",		US_isopen,	usersub, filename);
    make_usub("isread",		US_isread,	usersub, filename);
    make_usub("isrecover",	US_isrecover,	usersub, filename);	
    make_usub("isrelease",	US_isrelease,	usersub, filename);
    make_usub("isrename",	US_isrename,	usersub, filename);
    make_usub("isrewcurr",	US_isrewcurr,	usersub, filename);
    make_usub("isrewrec",	US_isrewrec,	usersub, filename);
    make_usub("isrewrite",	US_isrewrite,	usersub, filename);
    make_usub("isrollback",	US_isrollback,	usersub, filename);
    make_usub("issetunique",	US_issetunique,	usersub, filename);
    make_usub("isstart",	US_isstart,	usersub, filename);
    make_usub("isuniqueid",	US_isuniqueid,	usersub, filename);
    make_usub("isunlock",	US_isunlock,	usersub, filename);
    make_usub("iswrcurr",	US_iswrcurr,	usersub, filename);
    make_usub("iswrite",	US_iswrite,	usersub, filename);
/* format conversion */
    make_usub("ldchar",		US_ldchar,	usersub, filename);
    make_usub("lddbl",		US_lddbl,	usersub, filename);
    make_usub("lddblnull",	US_lddblnull,	usersub, filename);
    make_usub("lddecimal",	US_lddecimal,	usersub, filename);
    make_usub("ldfloat",	US_ldfloat,	usersub, filename);
    make_usub("ldfltnull",	US_ldfltnull,	usersub, filename);
    make_usub("ldint",		US_ldint,	usersub, filename);
    make_usub("ldlong",		US_ldlong,	usersub, filename);
    make_usub("stchar",		US_stchar,	usersub, filename);
    make_usub("stdbl",		US_stdbl,	usersub, filename);
    make_usub("stdblnull",	US_stdblnull,	usersub, filename);
    make_usub("stdecimal",	US_stdecimal,	usersub, filename);
    make_usub("stfloat",	US_stfloat,	usersub, filename);
    make_usub("stfltnull",	US_stfltnull,	usersub, filename);
    make_usub("stint",		US_stint,	usersub, filename);
    make_usub("stlong",		US_stlong,	usersub, filename);
/* decimal functions */
    make_usub("deccvasc",	US_deccvasc,	usersub, filename);
    make_usub("dectoasc",	US_dectoasc,	usersub, filename);
    make_usub("deccvint",	US_deccvint,	usersub, filename);
    make_usub("dectoint",	US_dectoint,	usersub, filename);
    make_usub("deccvlong",	US_deccvlong,	usersub, filename);
    make_usub("dectolong",	US_dectolong,	usersub, filename);
    make_usub("deccvflt",	US_deccvflt,	usersub, filename);
    make_usub("dectoflt",	US_dectoflt,	usersub, filename);
    make_usub("deccvdbl",	US_deccvdbl,	usersub, filename);
    make_usub("dectodbl",	US_dectodbl,	usersub, filename);
    make_usub("decadd",		US_decadd,	usersub, filename);
    make_usub("decsub",		US_decsub,	usersub, filename);
    make_usub("decmul",		US_decmul,	usersub, filename);
    make_usub("decdiv",		US_decdiv,	usersub, filename);
    make_usub("deccmp",		US_deccmp,	usersub, filename);
    make_usub("deccopy",	US_deccopy,	usersub, filename);
    make_usub("dececvt",	US_dececvt,	usersub, filename);
    make_usub("decfcvt",	US_decfcvt,	usersub, filename);
  };

  
static int
usersub(ix, sp, items)
int ix;
register int sp;
register int items;
{
    STR **st = stack->ary_array + sp;
    register int i;
    register char *tmps;
    register STR *Str;		/* used in str_get and str_gnum macros */

    switch (ix) {
/* c-isam file manipulation */
    case US_isaddindex:
#ifdef READONLY
fatal("&isaddindex:%s",ROMSG);
#else
	if (items != 2)
	    fatal("Usage: &isaddindex($isfd, $keydesc)");
	else {
	    int retval;
	    int		isfd =		(int)		str_gnum(st[1]);
	    struct keydesc *	keydesc =	(struct keydesc *)	str_get(st[2]);

	    retval = isaddindex(isfd, keydesc);
	    str_numset(st[0], (double) retval);
	}
	return sp;
#endif

    case US_isaudit:
#ifdef READONLY
fatal("&isaudit:%s",ROMSG);
#else
	if (items != 3)
	    fatal("Usage: &isaudit($isfd, $filename, $mode)");
	else {
	    int retval;
	    int		isfd =		(int)		str_gnum(st[1]);
	    char *	filename =	(char *)	str_get(st[2]);
	    int		mode =		(int)		str_gnum(st[3]);

	    STR_GROW(st[2],66); filename = str_get(st[2]);

	    retval = isaudit(isfd, filename, mode);
	    str_numset(st[0], (double) retval);
	    str_set(st[2], (char*) filename);
	}
	return sp;
#endif

    case US_isbegin:
	if (items != 0)
	    fatal("Usage: &isbegin()");
	else {
	    int retval;

	    retval = isbegin();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_isbuild:
#ifdef READONLY
fatal("&isbuild:%s",ROMSG);
#else
	if (items != 4)
	    fatal("Usage: &isbuild($filename, $reclen, $keydesc, $mode)");
	else {
	    char	*fn;
	    int retval;
	    char *	filename =	(char *)	str_get(st[1]);
	    int		reclen =	(int)		str_gnum(st[2]);
	    struct keydesc *	keydesc =	(struct keydesc *)	str_get(st[3]);
	    int		mode =		(int)		str_gnum(st[4]);
	    retval = isbuild(filename, reclen, keydesc, mode);
	    str_numset(st[0], (double) retval);

	    isam_reclen[retval] = reclen;
	}
	return sp;
#endif

    case US_isclose:
	if (items != 1)
	    fatal("Usage: &isclose($isfd)");
	else {
	    int retval;
	    int		isfd =		(int)		str_gnum(st[1]);

	    retval = isclose(isfd);
	    str_numset(st[0], (double) retval);

	    /*  forget record length for this file */

	    isam_reclen[isfd] == -1;
	}
	return sp;

    case US_iscluster:
#ifdef READONLY
fatal("&iscluster:%s",ROMSG);
#else
	if (items != 2)
	    fatal("Usage: &iscluster($isfd, $keydesc)");
	else {
	    int retval;
	    int		isfd =		(int)		str_gnum(st[1]);
	    struct keydesc *	keydesc =	(struct keydesc *)	str_get(st[2]);

	    retval = iscluster(isfd, keydesc);
	    str_numset(st[0], (double) retval);
	}
	return sp;
#endif

    case US_iscommit:
#ifdef READONLY
fatal("&iscommit:%s",ROMSG);
#else
	if (items != 0)
	    fatal("Usage: &iscommit()");
	else {
	    int retval;

	    retval = iscommit();
	    str_numset(st[0], (double) retval);
	}
	return sp;
#endif

    case US_isdelcurr:
#ifdef READONLY
fatal("&isdelcurr:%s",ROMSG);
#else
	if (items != 1)
	    fatal("Usage: &isdelcurr($isfd)");
	else {
	    int retval;
	    int		isfd =		(int)		str_gnum(st[1]);

	    retval = isdelcurr(isfd);
	    str_numset(st[0], (double) retval);
	}
	return sp;
#endif

    case US_isdelete:
#ifdef READONLY
fatal("&isdelete:%s",ROMSG);
#else
	if (items != 2)
	    fatal("Usage: &isdelete($isfd, $record)");
	else {
	    int retval;
	    int		isfd =		(int)		str_gnum(st[1]);
	    char *	record =	(char *)	str_get(st[2]);

	    retval = isdelete(isfd, record);
	    str_numset(st[0], (double) retval);
	}
	return sp;
#endif

    case US_isdelindex:
#ifdef READONLY
fatal("&isdelindex:%s",ROMSG);
#else
	if (items != 2)
	    fatal("Usage: &isdelindex($isfd, $keydesc)");
	else {
	    int retval;
	    int		isfd =		(int)		str_gnum(st[1]);
	    struct keydesc *	keydesc =	(struct keydesc *)	str_get(st[2]);

	    retval = isdelindex(isfd, keydesc);
	    str_numset(st[0], (double) retval);
	}
	return sp;
#endif

    case US_isdelrec:
#ifdef READONLY
fatal("&isdelrec:%s",ROMSG);
#else
	if (items != 2)
	    fatal("Usage: &isdelrec($isfd, $recnum)");
	else {
	    int retval;
	    int		isfd =		(int)		str_gnum(st[1]);
	    long	recnum =	(long)		str_gnum(st[2]);

	    retval = isdelrec(isfd, recnum);
	    str_numset(st[0], (double) retval);
	}
	return sp;
#endif

    case US_iserase:
#ifdef READONLY
fatal("&iserase:%s",ROMSG);
#else
	if (items != 1)
	    fatal("Usage: &iserase($filename)");
	else {
	    int retval;
	    char *	filename =	(char *)	str_get(st[1]);

	    retval = iserase(filename);
	    str_numset(st[0], (double) retval);
	}
	return sp;
#endif

    case US_isflush:
	if (items != 1)
	    fatal("Usage: &isflush($isfd)");
	else {
	    int retval;
	    int		isfd =		(int)		str_gnum(st[1]);

	    retval = isflush(isfd);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_isindexinfo:
	if (items != 3)
	    fatal("Usage: &isindexinfo($isfd, $buffer, $ixnumber)");
	else {
	    int retval;
	    int		isfd =		(int)		str_gnum(st[1]);
	    char *	buffer =	(char *)	str_get(st[2]);
	    int		ixnumber =	(int)		str_gnum(st[3]);
	    int		size;

	    if (ixnumber == 0) {
		size = sizeof(struct keydesc);
		}
	    else {
		size = sizeof(struct dictinfo);
		}
	    STR_GROW(st[2],size+1);
	    buffer = str_get(st[2]);

	    retval = isindexinfo(isfd, buffer, ixnumber);
	    str_numset(st[0], (double) retval);
	    str_nset(st[2], buffer, size+1);
	}
	return sp;

    case US_islock:
	if (items != 1)
	    fatal("Usage: &islock($isfd)");
	else {
	    int retval;
	    int		isfd =		(int)		str_gnum(st[1]);

	    retval = islock(isfd);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_islogclose:
#ifdef READONLY
fatal("&islogclose:%s",ROMSG);
#else
	if (items != 0)
	    fatal("Usage: &islogclose()");
	else {
	    int retval;

	    retval = islogclose();
	    str_numset(st[0], (double) retval);
	}
	return sp;
#endif

    case US_islogopen:
#ifdef READONLY
fatal("&islogopen:this is readlonly version of cisamperl");
#else
	if (items != 1)
	    fatal("Usage: &islogopen($logname)");
	else {
	    int retval;
	    char *	logname =	(char *)	str_get(st[1]);

	    retval = islogopen(logname);
	    str_numset(st[0], (double) retval);
	}
	return sp;
#endif

    case US_isopen:
	if (items != 2)
	    fatal("Usage: &isopen($filename, $mode)");
	else {
	    struct dictinfo tmp;
	    int retval;
	    char *	filename =	(char *)	str_get(st[1]);
	    int		mode =		(int)		str_gnum(st[2]);

#ifdef READONLY
	    if ((mode&0x0f) > 0) {
		fatal("&isopen:%s",ROMSG);
		}
#endif
	    retval = isopen(filename, mode);

	    /* get & remember the record length for this file */

	   if (isindexinfo(retval,&tmp,0) == -1) {
		fprintf(stderr,"iserrno == %d\n",iserrno);
		fatal("isopen cannot call isindexinfo for dictinfo");
		}
#ifdef DEBUG
fprintf(stderr,"remembering: isam_reclen[%d] = %d\n",retval,tmp.di_recsize);
#endif
	    isam_reclen[retval] = tmp.di_recsize;

	    str_numset(st[0], (double) retval);
	}
	return sp;


    case US_isread:
	if (items != 3)
	    fatal("Usage: &isread($isfd, $record, $mode)");
	else {
	    int retval;
	    int		isfd =		(int)		str_gnum(st[1]);
	    char	*record = 	(char	*)	str_get(st[2]);
	    int		mode =		(int)		str_gnum(st[3]);

		/*
		 * first grow the passed string to the required length
		 * which we remembered on open, then get the pointer
		 * again, cause STR_GROW might have put the whole
		 * thing somewhere else
		 * sneaky, stolen from O_SYSREAD in eval.c
		 */

#ifdef DEBUG
fprintf(stderr,"record = 0x%x\n",record);
fprintf(stderr,"   len = %d\n",(st[2])->str_len);
#endif
	    STR_GROW(st[2],isam_reclen[isfd]+1), record = str_get(st[2]);
#ifdef DEBUG
fprintf(stderr,"record = 0x%x\n",record);
fprintf(stderr,"   len = %d\n",(st[2])->str_len);
#endif

	    retval = isread(isfd, record, mode);
	    str_numset(st[0], (double) retval);
		/*
		 * do not str_set here, since that scans for
		 * \0, maybe str_nset?
		 */
	    str_nset(st[2], (char*) record,isam_reclen[isfd]+1);
	}
	return sp;

    case US_isrecover:
#ifdef READONLY
fatal("&isrecover:%s",ROMSG);
#else
	if (items != 0)
	    fatal("Usage: &isrecover()");
	else {
	    int retval;

	    retval = isrecover();
	    str_numset(st[0], (double) retval);
	}
	return sp;
#endif

    case US_isrelease:
	if (items != 1)
	    fatal("Usage: &isrelease($isfd)");
	else {
	    int retval;
	    int		isfd =		(int)		str_gnum(st[1]);

	    retval = isrelease(isfd);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_isrename:
#ifdef READONLY
fatal("&isrename:%s",ROMSG);
#else
	if (items != 2)
	    fatal("Usage: &isrename($oldname, $newname)");
	else {
	    int retval;
	    char *	oldname =	(char *)	str_get(st[1]);
	    char *	newname =	(char *)	str_get(st[2]);

	    retval = isrename(oldname, newname);
	    str_numset(st[0], (double) retval);
	}
	return sp;
#endif

    case US_isrewcurr:
#ifdef READONLY
fatal("&isrewcurr:%s",ROMSG);
#else
	if (items != 2)
	    fatal("Usage: &isrewcurr($isfd, $record)");
	else {
	    int retval;
	    int		isfd =		(int)		str_gnum(st[1]);
	    char *	record =	(char *)	str_get(st[2]);
	    char *	buffer;
	    int		count;

		/* allocate an EXACT buffer, copy the passed record
		 * information in, and pad with nulls
		 */
	    if ((buffer = malloc(isam_reclen[isfd]+1)) == NULL)
		{
		fatal("malloc error");
		}
		/* if the passed record is actually smaller than the
		 * record expected, we pas with zeros 
		 */
	    count = ((st[2])->str_len < isam_reclen[isfd] ?
			(st[2])->str_len:
			isam_reclen[isfd]);
	    memset(buffer,'\0',isam_reclen[isfd]+1);
	    memcpy(buffer,record,count);
  	    retval = isrewcurr(isfd, buffer);
	    free(buffer);

	    str_numset(st[0], (double) retval);
	}
	return sp;
#endif

    case US_isrewrec:
#ifdef READONLY
fatal("&isrewrec:%s",ROMSG);
#else
	if (items != 3)
	    fatal("Usage: &isrewrec($isfd, $recnum, $record)");
	else {
	    int retval;
	    int		isfd =		(int)		str_gnum(st[1]);
	    long	recnum =	(long)		str_gnum(st[2]);
	    char *	record =	(char *)	str_get(st[3]);
	    char *	buffer;
	    int		count;

		/* allocate an EXACT buffer, copy the passed record
		 * information in, and pad with nulls
		 */
	    if ((buffer = malloc(isam_reclen[isfd]+1)) == NULL)
		{
		fatal("malloc error");
		}
	    count = ((st[2])->str_len < isam_reclen[isfd] ?
			(st[2])->str_len :
			isam_reclen[isfd]);
	    memset(buffer,'\0',isam_reclen[isfd]+1);
	    memcpy(buffer,record,count);

	    retval = isrewrec(isfd, recnum, record);
	    free(buffer);

	    str_numset(st[0], (double) retval);
	}
	return sp;
#endif

    case US_isrewrite:
#ifdef READONLY
fatal("&isrewite:%s",ROMSG);
#else
	if (items != 2)
	    fatal("Usage: &isrewrite($isfd, $record)");
	else {
	    int retval;
	    int		isfd =		(int)		str_gnum(st[1]);
	    char *	record =	(char *)	str_get(st[2]);
	    char *	buffer;
	    int		count;

		/* allocate an EXACT buffer, copy the passed record
		 * information in, and pad with nulls
		 */
	    if ((buffer = malloc(isam_reclen[isfd]+1)) == NULL)
		{
		fatal("malloc error");
		}
	    count = ((st[2])->str_len < isam_reclen[isfd] ?
			(st[2])->str_len :
			isam_reclen[isfd]);
	    memset(buffer,'\0',isam_reclen[isfd]+1);
	    memcpy(buffer,record,count);

	    retval = isrewrite(isfd, record);
	    free(buffer);
	    str_numset(st[0], (double) retval);
	}
	return sp;
#endif

    case US_isrollback:
#ifdef READONLY
fatal("&isrollback:%s",ROMSG);
#else
	if (items != 0)
	    fatal("Usage: &isrollback()");
	else {
	    int retval;

	    retval = isrollback();
	    str_numset(st[0], (double) retval);
	}
	return sp;
#endif

    case US_issetunique:
#ifdef READONLY
fatal("&issetunique:%s",ROMSG);
#else
	if (items != 2)
	    fatal("Usage: &issetunique($isfd, $uniqueid)");
	else {
	    int retval;
	    int		isfd =		(int)		str_gnum(st[1]);
	    long	uniqueid =	(long)		str_gnum(st[2]);

	    retval = issetunique(isfd, uniqueid);
	    str_numset(st[0], (double) retval);
	}
	return sp;
#endif

    case US_isstart:
	if (items != 5)
	    fatal("Usage: &isstart($isfd, $keydesc, $length, $record, $mode)");
	else {
	    int 	retval;
	    int		isfd =		(int)		str_gnum(st[1]);
	    struct keydesc *	keydesc =	(struct keydesc *)	str_get(st[2]);
	    int		length =	(int)		str_gnum(st[3]);
	    char *	record =	(char *)	str_get(st[4]);
	    int		mode =		(int)		str_gnum(st[5]);

		/* grow the string to needed space */
	    STR_GROW(st[4],isam_reclen[isfd]+1); 
		record = (char *) str_get(st[4]);
	    STR_GROW(st[2],sizeof (struct keydesc)+1); 
		keydesc = (struct keydesc *) str_get(st[2]);

	    retval = isstart(isfd, keydesc, length, record, mode);


#ifdef DEBUG
	fprintf(stderr,"retval = %d %d %d\n",retval,iserrno,errno);
#endif
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_isuniqueid:
#ifdef READONLY
fatal("&isuniqueid:%s",ROMSG);
#else
	if (items != 2)
	    fatal("Usage: &isuniqueid($isfd, $uniqueid)");
	else {
	    int retval;
	    int		isfd =		(int)		str_gnum(st[1]);
	    long	uniqueid;

	    retval = isuniqueid(isfd, &uniqueid);
	    str_numset(st[0], (double) retval);
	    str_numset(st[2], (double) uniqueid);
	}
	return sp;
#endif

    case US_isunlock:
	if (items != 1)
	    fatal("Usage: &isunlock($isfd)");
	else {
	    int retval;
	    int		isfd =		(int)		str_gnum(st[1]);

	    retval = isunlock(isfd);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_iswrcurr:
#ifdef READONLY
fatal("&iswrcurr:%s",ROMSG);
#else
	if (items != 2)
	    fatal("Usage: &iswrcurr($isfd, $record)");
	else {
	    int retval;
	    int		isfd =		(int)		str_gnum(st[1]);
	    char *	record =	(char *)	str_get(st[2]);
	    char *	buffer;
	    int		count;

		/* allocate an EXACT buffer, copy the passed record
		 * information in, and pad with nulls
		 */
	    if ((buffer = malloc(isam_reclen[isfd]+1)) == NULL)
		{
		fatal("malloc error");
		}
	    count = ((st[2])->str_len < isam_reclen[isfd] ?
			(st[2])->str_len :
			isam_reclen[isfd]);
	    memset(buffer,'\0',isam_reclen[isfd]+1);
	    memcpy(buffer,record,count);

	    retval = iswrcurr(isfd, record);
	    free(buffer);
	    str_numset(st[0], (double) retval);
	}
	return sp;
#endif

    case US_iswrite:
#ifdef READONLY
fatal("&iswrite:%s",ROMSG);
#else
	if (items != 2)
	    fatal("Usage: &iswrite($isfd, $record)");
	else {
	    int retval;
	    int		isfd =		(int)		str_gnum(st[1]);
	    char *	record =	(char *)	str_get(st[2]);
	    char *	buffer;
	    int		count;

		/* allocate an EXACT buffer, copy the passed record
		 * information in, and pad with nulls
		 */
	    if ((buffer = malloc(isam_reclen[isfd]+1)) == NULL)
		{
		fatal("malloc error");
		}
	    count = ((st[2])->str_len < isam_reclen[isfd] ?
			(st[2])->str_len :
			isam_reclen[isfd]);
	    memset(buffer,'\0',isam_reclen[isfd]+1);
	    memcpy(buffer,record,count);

	    retval = iswrite(isfd, record);
	    str_numset(st[0], (double) retval);
	}
	return sp;
#endif

/* format conversion */
    case US_ldchar:
	if (items != 4)
	    fatal("Usage: &ldchar($fstr, $offset, $length, $cstr)");
	else {
	    int retval;
	    char *	fstr =		(char *)	str_get(st[1]);
	    int		offset =	(int)		str_gnum(st[2]);
	    int		length =	(int)		str_gnum(st[3]);
	    char	*cstr = 	(char *)	str_get(st[4]);

	    if (offset > (((st[1])->str_len)-length)) {
		fatal("&ldchar overshoots end of buffer");
	    }
		/* grow the string to needed space */
	    STR_GROW(st[4],length+1); cstr = str_get(st[4]);

	    retval = ldchar(fstr, length, cstr);
	    str_numset(st[0], (double) retval);
	    str_set(st[4], (char*) cstr);
	}
	return sp;

    case US_lddbl:
	if (items != 2)
	    fatal("Usage: &lddbl($p, $offset)");
	else {
	    double retval;
	    char *	p =		(char *)	str_get(st[1]);
	    int		offset =	(int)		str_gnum(st[2]);

	    if (offset > (((st[1])->str_len)-DOUBLESIZE)) {
		fatal("&lddbl overshoots end of buffer");
	    }
	    retval = lddbl(p+offset);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_lddblnull:
	if (items != 3)
	    fatal("Usage: &lddblnull($p, $offset, $nullflag)");
	else {
	    double retval;
	    char *	p =		(char *)	str_get(st[1]);
	    int		offset =	(int)		str_gnum(st[2]);
	    short	nullflag;

	    if (offset > (((st[1])->str_len)-DOUBLESIZE)) {
		fatal("&lddblnull overshoots end of buffer");
	    }
	    retval = lddblnull(p+offset, &nullflag);
	    str_numset(st[0], (double) retval);
	    str_numset(st[3], (double) nullflag);
	}
	return sp;

    case US_lddecimal:
#ifdef NODECIMALS
	fatal(NODECMSG);
#else
	if (items != 4)
	    fatal("Usage: &lddecimal($cp, $offset, $len, $decp)");
	else {
	    int retval;
	    char *	cp =		(char *)	str_get(st[1]);
	    int		offset =	(int)		str_gnum(st[2]);
	    int		len =		(int)		str_gnum(st[3]);
	    char *	decp =		(char *)	str_get(st[4]);

	    if (offset > (((st[1])->str_len)-len)) {
		fatal("&lddecimal overshoots end of buffer");
	    }
	
	    STR_GROW(st[4],sizeof(dec_t)+1); decp = str_get(st[4]);

	    retval = lddecimal(cp+offset, len, (dec_t *)decp);
	    str_numset(st[0], (double) retval);
	    str_nset(st[4], (char*) decp, sizeof(dec_t)+1);
	}
	return sp;
#endif

    case US_ldfloat:
	if (items != 2)
	    fatal("Usage: &ldfloat($p, $offset)");
	else {
	    double retval;
	    char *	p =		(char *)	str_get(st[1]);
	    int		offset =	(int)		str_gnum(st[2]);

	    if (offset > (((st[1])->str_len)-FLOATSIZE)) {
		fatal("&ldfloat overshoots end of buffer");
	    }
	    retval = ldfloat(p+offset);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_ldfltnull:
	if (items != 3)
	    fatal("Usage: &ldfltnull($p, $offset, $nullflag)");
	else {
	    int retval;
	    char *	p =		(char *)	str_get(st[1]);
	    int		offset =	(int)		str_gnum(st[2]);
	    short	nullflag;

	    if (offset > (((st[1])->str_len)-FLOATSIZE)) {
		fatal("&ldfltnull overshoots end of buffer");
	    }
	    retval = ldfltnull(p+offset, &nullflag);
	    str_numset(st[0], (double) retval);
	    str_numset(st[3], (double) nullflag);
	}
	return sp;

    case US_ldint:
	if (items != 2)
	    fatal("Usage: &ldint($p, $offset)");
	else {
	    int retval;
	    char *	p =		(char *)	str_get(st[1]);
	    int		offset =	(int)		str_gnum(st[2]);

	    if (offset > (((st[1])->str_len)-INTSIZE)) {
		fatal("&ldint overshoots end of buffer");
	    }
	    retval = ldint(p+offset);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_ldlong:
	if (items != 2)
	    fatal("Usage: &ldlong($p, $offset)");
	else {
	    long retval;
	    char *	p =		(char *)	str_get(st[1]);
	    int		offset =	(int)		str_gnum(st[2]);

	    if (offset > (((st[1])->str_len)-LONGSIZE)) {
		fatal("&ldint overshoots end of buffer");
	    }
	    retval = ldlong(p+offset);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_stchar:
	if (items != 4)
	    fatal("Usage: &stchar($cstr, $fstr, $offset, $length)");
	else {
	    int retval;
	    char *	cstr =		(char *)	str_get(st[1]);
	    char *	fstr =		(char *)	str_get(st[2]);
	    int		offset =	(int)		str_gnum(st[3]);
	    int		length =	(int)		str_gnum(st[4]);

	    if ((offset + length+1) > (st[2])->str_len) {
		STR_GROW(st[2],offset+length+1);
		fstr = str_get(st[2]);
		}

	    retval = stchar(cstr, fstr+offset, length);
	    str_numset(st[0], (double) retval);
	    str_nset(st[2], (char*) fstr,(st[2])->str_len);
	}
	return sp;

    case US_stdbl:
	if (items != 3)
	    fatal("Usage: &stdbl($d, $p, $offset)");
	else {
	    int retval;
	    double	d =		(double)	str_gnum(st[1]);
	    char *	p =		(char *)	str_get(st[2]);
	    int		offset =	(int)		str_gnum(st[3]);

	    if ((offset + DOUBLESIZE+1) > (st[2])->str_len) {
		STR_GROW(st[2],offset+DOUBLESIZE+1);
		p = str_get(st[2]);
		}
	
	    retval = stdbl(d, p+offset);
	    str_numset(st[0], (double) retval);
	    str_nset(st[2], (char*) p,(st[2])->str_len);
	}
	return sp;

    case US_stdblnull:
	if (items != 4)
	    fatal("Usage: &stdblnull($d, $p, $offset, $nullflag)");
	else {
	    int retval;
	    double	d =		(double)	str_gnum(st[1]);
	    char *	p =		(char *)	str_get(st[2]);
	    int		offset =	(int)		str_gnum(st[3]);
	    short	nullflag;

	    if ((offset + DOUBLESIZE+1) > (st[2])->str_len) {
		STR_GROW(st[2],offset+DOUBLESIZE+1);
		p = str_get(st[2]);
		}

	    retval = stdblnull(d, p+offset, &nullflag);
	    str_numset(st[0], (double) retval);
	    str_nset(st[2], (char*) p,(st[2])->str_len);
	    str_numset(st[4], (double) nullflag);
	}
	return sp;

    case US_stdecimal:
#ifdef NODECIMAL
	fatal(NODECMSG);
#else
	if (items != 4)
	    fatal("Usage: &stdecimal($decp, $cp, $offset, $len)");
	else {
	    int retval;
	    char *	decp =		(char *)	str_get(st[1]);
	    char *	cp = 		(char *)	str_get(st[2]);
	    int		offset =	(int)		str_gnum(st[3]);
	    int		len =		(int)		str_gnum(st[4]);

		
	    if ((offset + len+1) > (st[2])->str_len) {
		STR_GROW(st[2],offset+len+1);
		cp = str_get(st[2]);
		}

	    retval = stdecimal((dec_t *)decp, cp+offset, len);
	    str_numset(st[0], (double) retval);
	    str_nset(st[2], (char*) cp,(st[2])->str_len);
	}
	return sp;
#endif

    case US_stfloat:
	if (items != 3)
	    fatal("Usage: &stfloat($f, $p, $offset)");
	else {
	    int retval;
	    float	f =		(float)		str_gnum(st[1]);
	    char *	p =		(char *)	str_get(st[2]);
	    int		offset =	(int)		str_gnum(st[3]);

	    if ((offset + FLOATSIZE+1) > (st[2])->str_len) {
		STR_GROW(st[2],offset+FLOATSIZE+1);
		p = str_get(st[2]);
		}

	    retval = stfloat(f, p+offset);
	    str_numset(st[0], (double) retval);
	    str_nset(st[2], (char*) p,(st[2])->str_len);
	}
	return sp;

    case US_stfltnull:
	if (items != 4)
	    fatal("Usage: &stfltnull($f, $p, $offset, $nlflg)");
	else {
	    int retval;
	    float	f =		(float)		str_gnum(st[1]);
	    char *	p =		(char *)	str_get(st[2]);
	    int		offset =	(int)		str_gnum(st[3]);
	    short	nlflg;

	    if ((offset + FLOATSIZE+1) > (st[2])->str_len) {
		STR_GROW(st[2],offset+FLOATSIZE+1);
		p = str_get(st[2]);
		}

	    retval = stfltnull(f, p+offset, &nlflg);
	    str_numset(st[0], (double) retval);
	    str_nset(st[2], (char*) p,(st[2])->str_len);
	    str_numset(st[3], (double) nlflg);
	}
	return sp;

    case US_stint:
	if (items != 3)
	    fatal("Usage: &stint($i, $p, $offset)");
	else {
	    int retval;
	    int		i =		(int)		str_gnum(st[1]);
	    char *	p =		(char *)	str_get(st[2]);
	    int		offset =	(int)		str_gnum(st[3]);

	    if ((offset + INTSIZE+1) > (st[2])->str_len) {
		STR_GROW(st[2],offset+INTSIZE+1);
		p = str_get(st[2]);
		}

	    retval = stint(i, p+offset);
	    str_numset(st[0], (double) retval);
	    str_nset(st[2], (char*) p,(st[2])->str_len);
	}
	return sp;

    case US_stlong:
	if (items != 3)
	    fatal("Usage: &stlong($l, $p, $offset)");
	else {
	    int retval;
	    long	l =		(long)		str_gnum(st[1]);
	    char *	p =		(char *)	str_get(st[2]);
	    int		offset =	(int)		str_gnum(st[3]);

	    if ((offset + LONGSIZE+1) > (st[2])->str_len) {
		STR_GROW(st[2],offset+LONGSIZE+1);
		p = str_get(st[2]);
		}

	    retval = stlong(l, p+offset);
	    str_numset(st[0], (double) retval);
	    str_nset(st[2], (char*) p,(st[2])->str_len);
	}
	return sp;

/* decimal functions */
    case US_deccvasc:
#ifdef NODECIMAL
	fatal(NODECMSG);
#else
	if (items != 4)
	    fatal("Usage: &deccvasc($cp, $offset, $len, $np)");
	else {
	    int retval;
	    char *	cp =		(char *)	str_get(st[1]);
	    int		offset =	(int)		str_gnum(st[2]);
	    int		len =		(int)		str_gnum(st[3]);
	    char *	np =		(char *)	str_get(st[4]);

	    STR_GROW(st[4],sizeof(dec_t)+1); np = str_get(st[4]);

	    retval = deccvasc(cp+offset, len, (dec_t *)np);
	    str_numset(st[0], (double) retval);
	    str_nset(st[4], (char*) np,sizeof(dec_t)+1);
	}
	return sp;
#endif

    case US_dectoasc:
#ifdef NODECIMAL
	fatal(NODECMSG);
#else
	if (items != 5)
	    fatal("Usage: &dectoasc($np, $cp, $offset, $len, $right)");
	else {
	    int retval;
	    char *	np =		(char *)	str_get(st[1]);
	    char *	cp =		(char *)	str_get(st[2]);
	    int		offset =	(int)		str_gnum(st[3]);
	    int		len =		(int)		str_gnum(st[4]);
	    int		right =		(int)		str_gnum(st[5]);

	    if ((offset + len+1) > ((st[2])->str_len)) {
		STR_GROW(st[2],offset+len+1);
		cp = str_get(st[2]);
		}

	    retval = dectoasc((dec_t *)np, cp+offset, len, right);
	    str_numset(st[0], (double) retval);
	    str_set(st[2], (char*) cp);
	}
	return sp;
#endif

    case US_deccvint:
#ifdef NODECIMAL
	fatal(NODECMSG);
#else
	if (items != 2)
	    fatal("Usage: &deccvint($integer, $np)");
	else {
	    int retval;
	    int		integer =	(int)		str_gnum(st[1]);
	    char *	np =		(char *)		str_get(st[2]);

	    STR_GROW(st[2],sizeof(dec_t)+1); np = str_get(st[2]);

	    retval = deccvint(integer, (dec_t *)np);
	    str_numset(st[0], (double) retval);
	    str_nset(st[2], (char*) np,sizeof(dec_t)+1);
	}
	return sp;
#endif

    case US_dectoint:
#ifdef NODECIMAL
	fatal(NODECMSG);
#else
	if (items != 2)
	    fatal("Usage: &dectoint($np, $integer)");
	else {
	    int retval;
	    char *	np =		(char *)	str_get(st[1]);
	    int		integer;

 	    retval = dectoint((dec_t *)np, &integer);
	    str_numset(st[0], (double) retval);
	    str_numset(st[2], (double) integer);
	}
	return sp;
#endif

    case US_deccvlong:
#ifdef NODECIMAL
	fatal(NODECMSG);
#else
	if (items != 2)
	    fatal("Usage: &deccvlong($lng, $np)");
	else {
	    int retval;
	    long	lng =		(long)		str_gnum(st[1]);
	    char *	np =		(char *)	str_get(st[2]);

	    STR_GROW(st[2],sizeof(dec_t)+1); np = str_get(st[2]);

	    retval = deccvlong(lng, (dec_t *)np);
	    str_numset(st[0], (double) retval);
	    str_nset(st[2], (char*) np, sizeof(dec_t)+1);
	}
	return sp;
#endif

    case US_dectolong:
#ifdef NODECIMAL
	fatal(NODECMSG);
#else
	if (items != 2)
	    fatal("Usage: &dectolong($np, $lng)");
	else {
	    int retval;
	    char *	np =		(char *)	str_get(st[1]);
	    long	lng;

	    retval = dectolong((dec_t *)np, &lng);
	    str_numset(st[0], (double) retval);
	    str_numset(st[2], (double) lng);
	}
	return sp;
#endif

    case US_deccvflt:
#ifdef NODECIMAL
	fatal(NODECMSG);
#else
	if (items != 2)
	    fatal("Usage: &deccvflt($flt, $np)");
	else {
	    int retval;
	    float	flt =		(float)		str_gnum(st[1]);
	    char *	np =		(char *)	str_get(st[2]);

	    STR_GROW(st[2],sizeof(dec_t)+1); np = str_get(st[2]);

	    retval = deccvflt(flt, (dec_t *)np);
	    str_numset(st[0], (double) retval);
	    str_nset(st[2], (char*) np, sizeof(dec_t)+1);
	}
	return sp;
#endif

    case US_dectoflt:
#ifdef NODECIMAL
	fatal(NODECMSG);
#else
	if (items != 2)
	    fatal("Usage: &dectoflt($np, $flt)");
	else {
	    int retval;
	    char *	np =		(char *)	str_get(st[1]);
	    float	flt;

	    retval = dectoflt((dec_t *)np, &flt);
	    str_numset(st[0], (double) retval);
	    str_numset(st[2], (double) flt);
	}
	return sp;
#endif

    case US_deccvdbl:
#ifdef NODECIMAL
	fatal(NODECMSG);
#else
	if (items != 2)
	    fatal("Usage: &deccvdbl($dbl, $np)");
	else {
	    int retval;
	    double	dbl =		(double)	str_gnum(st[1]);
	    char *	np =		(char *)	str_get(st[2]);

	    STR_GROW(st[2],sizeof(dec_t)+1); np = str_get(st[2]);

	    retval = deccvdbl(dbl, (dec_t *)np);
	    str_numset(st[0], (double) retval);
	    str_nset(st[2], (char*) np, sizeof(dec_t)+1);
	}
	return sp;
#endif

    case US_dectodbl:
#ifdef NODECIMAL
	fatal(NODECMSG);
#else
	if (items != 2)
	    fatal("Usage: &dectodbl($np, $dbl)");
	else {
	    int retval;
	    char *	np =		(char *)	str_get(st[1]);
	    double	dbl;

	    retval = dectodbl((dec_t *)np, &dbl);
	    str_numset(st[0], (double) retval);
	    str_numset(st[2], (double) dbl);
	}
	return sp;
#endif

    case US_decadd:
#ifdef NODECIMAL
	fatal(NODECMSG);
#else
	if (items != 3)
	    fatal("Usage: &decadd($n1, $n2, $result)");
	else {
	    int retval;
	    char *	n1 =		(char *)	str_get(st[1]);
	    char *	n2 =		(char *)	str_get(st[2]);
	    char *	result =	(char *)	str_get(st[3]);

	    STR_GROW(st[3],sizeof(dec_t)+1); result = str_get(st[3]);

	    retval = decadd((dec_t *)n1, (dec_t *)n2, (dec_t *)result);
	    str_numset(st[0], (double) retval);
	    str_nset(st[3], (char*) result, sizeof(dec_t)+1);
	}
	return sp;
#endif

    case US_decsub:
#ifdef NODECIMAL
	fatal(NODECMSG);
#else
	if (items != 3)
	    fatal("Usage: &decsub($n1, $n2, $result)");
	else {
	    int retval;
	    char *	n1 =		(char *)	str_get(st[1]);
	    char *	n2 =		(char *)	str_get(st[2]);
	    char *	result =	(char *)	str_get(st[3]);

	    STR_GROW(st[3],sizeof(dec_t)+1); result = str_get(st[3]);

	    retval = decsub((dec_t *)n1, (dec_t *)n2, (dec_t *)result);
	    str_numset(st[0], (double) retval);
	    str_nset(st[3], (char*) result, sizeof(dec_t)+1);
	}
	return sp;
#endif

    case US_decmul:
#ifdef NODECIMAL
	fatal(NODECMSG);
#else
	if (items != 3)
	    fatal("Usage: &decmul($n1, $n2, $result)");
	else {
	    int retval;
	    char *	n1 =		(char *)	str_get(st[1]);
	    char *	n2 =		(char *)	str_get(st[2]);
	    char *	result =	(char *)	str_get(st[3]);

	    STR_GROW(st[3],sizeof(dec_t)+1); result = str_get(st[3]);

	    retval = decmul((dec_t *)n1, (dec_t *)n2, (dec_t *)result);
	    str_numset(st[0], (double) retval);
	    str_nset(st[3], (char*) result, sizeof(dec_t)+1);
	}
	return sp;
#endif

    case US_decdiv:
#ifdef NODECIMAL
	fatal(NODECMSG);
#else
	if (items != 3)
	    fatal("Usage: &decdiv($n1, $n2, $result)");
	else {
	    int retval;
	    char *	n1 =		(char *)	str_get(st[1]);
	    char *	n2 =		(char *)	str_get(st[2]);
	    char *	result =	(char *)	str_get(st[3]);

	    STR_GROW(st[3],sizeof(dec_t)+1); result = str_get(st[3]);

	    retval = decdiv((dec_t *)n1, (dec_t *)n2, (dec_t *)result);
	    str_numset(st[0], (double) retval);
	    str_nset(st[3], (char*) result, sizeof(dec_t)+1);
	}
	return sp;
#endif

    case US_deccmp:
#ifdef NODECIMAL
	fatal(NODECMSG);
#else
	if (items != 2)
	    fatal("Usage: &deccmp($n1, $n2)");
	else {
	    int retval;
	    char *	n1 =		(char *)	str_get(st[1]);
	    char *	n2 =		(char *)	str_get(st[2]);

	    retval = deccmp((dec_t *)n1, (dec_t *)n2);
	    str_numset(st[0], (double) retval);
	}
	return sp;
#endif

    case US_deccopy:
	/* this could more easily be done by just copying the
	 * perl variable, but what the heck
	 */
#ifdef NODECIMAL
	fatal(NODECMSG);
#else
	if (items != 2)
	    fatal("Usage: &deccopy($n1, $n2)");
	else {
	    int retval;
	    char *	n1 =		(char *)	str_get(st[1]);
	    char *	n2 =		(char *)	str_get(st[2]);

	    STR_GROW(st[2],sizeof(dec_t)+1); n2 = str_get(st[2]);

	    retval = deccopy((dec_t *)n1, (dec_t *)n2);
	    str_numset(st[0], (double) retval);
	    str_nset(st[2], (char*) n2, sizeof(dec_t)+1);
	}
	return sp;
#endif

    case US_dececvt:
#ifdef NODECIMAL
	fatal(NODECMSG);
#else
	if (items != 4)
	    fatal("Usage: &dececvt($np, $ndigit, $decpt, $sign)");
	else {
	    char *	retval;
	    char *	np =		(char *)	str_get(st[1]);
	    int		ndigit =	(int)		str_gnum(st[2]);
	    int		decpt;
	    int		sign;

	    retval = (char *)dececvt((dec_t *)np, ndigit, &decpt, &sign);

	    str_set(st[0], retval);
	    str_numset(st[3], (double) decpt);

	    str_numset(st[4], (double) sign);
	}
	return sp;
#endif

    case US_decfcvt:
#ifdef NODECIMAL
	fatal(NODECMSG);
#else
	if (items != 4)
	    fatal("Usage: &decfcvt($np, $ndigit, $decpt, $sign)");
	else {
	    char *	retval;
	    char *	np =		(char *)	str_get(st[1]);
	    int		ndigit =	(int)		str_gnum(st[2]);
	    int		decpt;
	    int		sign;

	    retval = (char *)decfcvt((dec_t *)np, ndigit, &decpt, &sign);
	    str_set(st[0], retval);
	    str_numset(st[3], (double) decpt);

	    str_numset(st[4], (double) sign);
	}
	return sp;
#endif


    default:
	fatal("Unimplemented user-defined subroutine");
    }
    return sp;
}

static int
userval(ix, str)
int ix;
STR *str;
{
    switch (ix) {
    case UV_iserrno:
	str_numset(str, (double)iserrno);
	break;
    case UV_isrecnum:
	str_numset(str,(double)isrecnum);
	break;
    case UV_isstat1:
	str_numset(str,(double)isstat1);
	break;
    case UV_isstat2:
	str_numset(str,(double)isstat2);
	break;

    }
    return 0;
}

static int
userset(ix, str)
int ix;
STR *str;
{
    switch (ix) {
    case UV_iserrno:
	iserrno = (int)str_gnum(str);
	break;
    case UV_isrecnum:
	isrecnum = (int)str_gnum(str);
	break;
    case UV_isstat1:
	isstat1 = (int)str_gnum(str);
	break;
    case UV_isstat2:
	isstat2 = (int)str_gnum(str);
	break;
    }
    return 0;
}
